use std::collections::HashMap;

use serde::{Deserialize, Serialize};
use serde_json::json;

use crate::get_pg;
use crate::transaction::Transaction;
use crate::variant::Variant;

// ITEM
// VARIANTS
// QUANTIZATION
// IMPORT / EXPORT
// DEPENDENCIES
// DEMAND STATS
// SEASONAL REVIEWS

/// Represents a single item in a collection of physical goods.
///
/// This struct serves as a blueprint for describing individual items within
/// a larger inventory or database of physical goods. It includes fields for
/// the name of the item and its category.
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Item {
    /// The ID
    pub id: String,
    /// The name of the Item
    pub name: String,
    /// Category of the Item
    pub category: Option<String>,
    /// Image
    pub image_path: Option<String>,
    /// The variants of an Item.
    /// Each key of the `HashMap<_>` is the ID of a variant and contains a `Variant`
    pub variants: HashMap<String, Variant>,
}

pub fn get_image(path: &std::path::Path) -> Option<String> {
    let parent = path.parent()?;
    let file_name = path.file_stem()?.to_str()?;

    for ext in ["jpg", "jpeg", "webp", "png"] {
        let mut img_file = parent.to_path_buf();
        img_file.push(&format!("{file_name}.{ext}"));

        if img_file.exists() {
            return Some(img_file.display().to_string());
        }
    }

    None
}

impl Item {
    /// Creates a new `Item` from a parsed markdown document
    pub fn new(doc: &mdq::Document) -> Self {
        let path = std::path::Path::new(&doc.path);

        let id = path.file_stem().unwrap().to_str().unwrap().to_lowercase();

        let image_path = get_image(path);

        let category = doc
            .frontmatter
            .as_mapping()
            .unwrap()
            .get("category")
            .map(|x| x.as_str().unwrap().to_string());

        let name = doc
            .frontmatter
            .as_mapping()
            .unwrap()
            .get("name")
            .unwrap()
            .as_str()
            .unwrap()
            .to_string();

        let mut variants = HashMap::new();

        let variants_lst = doc
            .frontmatter
            .as_mapping()
            .unwrap()
            .get("variants")
            .map(|x| x.as_mapping().unwrap().clone())
            .unwrap();

        assert!(
            !variants_lst.is_empty(),
            "Item {id} needs at least one variant."
        );

        for (variant_id, variant) in variants_lst {
            variants.insert(
                variant_id.as_str().unwrap().to_string(),
                Variant::from_yml(&variant, variant_id.as_str().unwrap(), &id),
            );
        }

        Self {
            id,
            name,
            category,
            image_path,
            variants,
        }
    }
}

impl Item {
    /// Get this items variant names
    pub fn get_variants(&self) -> Vec<String> {
        self.variants.keys().cloned().collect()
    }

    /// Get a `Variant` of an `Item`
    pub fn variant(&self, variant: &str) -> Option<Variant> {
        self.variants.get(variant).cloned()
    }

    pub async fn inventory(&self) -> Vec<Transaction> {
        sqlx::query_as("SELECT * FROM transactions WHERE item = $1 AND consumed_timestamp IS NULL ORDER BY created DESC")
        .bind(&self.id)
        .fetch_all(get_pg!()).await.unwrap()
    }

    pub async fn inventory_by_origin(&self, origin: &str) -> Vec<Transaction> {
        sqlx::query_as("SELECT * FROM transactions WHERE item = $1 AND consumed_timestamp IS NULL AND origin = $2 ORDER BY created DESC")
        .bind(&self.id)
        .bind(origin)
        .fetch_all(get_pg!()).await.unwrap()
    }

    pub async fn consumed_by_destination(&self, destination: &str) -> Vec<Transaction> {
        sqlx::query_as("SELECT * FROM transactions WHERE item = $1 AND consumed_timestamp IS NOT NULL AND destination = $2 ORDER BY created DESC")
        .bind(&self.id)
        .bind(destination)
        .fetch_all(get_pg!()).await.unwrap()
    }

    pub fn api_json(&self) -> serde_json::Value {
        let variants: HashMap<String, serde_json::Value> = self
            .variants
            .iter()
            .map(|(key, value)| (key.clone(), value.api_json()))
            .collect();

        json!({
            "uuid": self.id,
            "image": if self.image_path.is_some() {
                Some(format!("/item/{}/image", self.id))
            } else {
                None
            },
            "name": self.name,
            "category": self.category,
            "variants": variants
        })
    }
}
